module Checkers
  VERSION = '0.0.0'

  module Sugar
    def not; @inverted = !inverted; self; end
    def is;  self; end
    def are; self; end
  end

  module Checks
    def empty
      check subject.empty?, "#{subject.inspect} should be empty, but isn't"
    end

    def ==(o)
      check subject == o, "#{subject.inspect} should be equal to #{o.inspect}, but isn't"
    end
    alias :equals :==

    def matches(pattern)
      check subject.match(pattern), "#{subject.inspect} should match with #{pattern.inspect}, but doesn't"
    end

    def includes(o)
      check subject.include?(o), "#{subject.inspect} should include #{o.inspect}, but doesn't"
    end

    def raises(e)
      subject.call
      check false,  "#{subject.inspect} should raise #{e.inspect}, doesn't"
    rescue e
      check true,  "#{subject.inspect} should raise #{e.inspect}, doesn't"
    end
  end


  class Check
    include Checks, Sugar
    attr_reader :subject, :caller, :message, :inverted, :run

    def initialize(subject, caller, run = Checkers.run)
      @subject = subject
      @caller = caller
      @message = ''
      @inverted = false
      @run = run
    end

    def check(condition, msg = 'Check failed.')
      self.message = msg
      condition = !condition if inverted
      condition ? run.pass(self) : run.fail(self)
    end

    private
    attr_writer :message
  end


  class Checklist
    attr_reader :name, :caller, :block, :error, :printer, :run

    def initialize(name, caller, printer = Checkers.printer, run = Checkers.run, &block)
      @name = name
      @caller = caller
      @block = block
      @printer = printer
      @run = run
    end

    def verify
      printer.checklist self
      instance_eval(&block)
    rescue StandardError => e
      @error = e
      run.fail(self)
    end

    def check(subject)
      Check.new(subject, caller)
    end
  end


  class Printer
    COLORS = { red: 31, green: 32, yellow: 33, blue: 34, magenta: 35, cyan: 36, gray:  90 }

    def color(color, string)
      "\e[#{COLORS[color]}m#{string}\e[0m"
    end

    COLORS.each do |name, _|
      define_method name, ->(string) { color(name, string) }
    end

    def pass(check)
      STDOUT.print green '+'
    end

    def fail(check)
      STDOUT.print red "F"
    end

    def checklist(checklist)
      name = checklist.name ? blue(checklist.name) : blue(checklist.caller.first.split(":in").first)
      STDOUT.print "\nChecking #{name}:  "
    end

    def summary passes, failures
      failed = failures.count
      passed = passes.count
      total  = failed + passed

      STDOUT.puts "\n\n\n\n#{blue total} checks verified:  #{green passed} Passed   #{red failed} Failed\n\n"

        failures.each.with_index do  |check, i|
          STDOUT.print red("#{i}) ")

          if check.is_a? Check
            STDOUT.puts "#{check.message}" << gray(" (#{check.caller.first})")
          elsif check.is_a? Checklist
            name = check.name ? "'#{check.name}'" : ""
            trace = format_trace(check.error.backtrace)
            STDOUT.puts "In checklist #{ blue "#{name}" } #{ gray "(#{check.caller.first})" }: #{red check.error.message} \n\n#{trace}\n\n"
          end
        end
        STDOUT.puts green "All good!" if failures.empty? && total > 0
    end

    def format_trace(trace)
      trace.join("\n")
           .gsub(":in", " ")
           .gsub(/`(.*)'/, "\\1") # method name
           .gsub(/:(\d+)/, gray(":\\1")) # line number
    end
  end


  class Run
    attr_reader :failures, :passes, :printer

    def initialize(printer)
      @failures = []
      @passes   = []
      @printer  = printer
    end

    def pass(check)
      passes << check
      printer.pass check
    end

    def fail(check)
      failures << check
      printer.fail check
    end

    def failed?
      !failures.empty?
    end

    def status
      failed? ? 0 : 1
    end

    def summary
      printer.summary passes, failures
    end
  end

  ## module methods
  def self.before_blocks
    @@before_blocks ||= []
  end

  def self.after_blocks
    @@after_blocks ||= []
  end

  def self.before_checking(&block)
    before_blocks << block
  end

  def self.after_checking(&block)
    after_blocks << block
  end

  def self.run
    @@run ||= Run.new printer
  end

  def self.printer=(printer)
    @@printer = printer
  end

  def self.printer
    @@printer ||= Printer.new
  end

  # extension
  module Helper
    def checklist(name = nil, &block)
      Checkers.before_blocks.each(&:call)
      Checklist.new(name, caller, &block).verify
      Checkers.after_blocks.each(&:call)
    end
  end

  extend Helper

  at_exit {
    Checkers.run.summary
    exit Checkers.run.status
  }
end
